<?php

/**
 * 克隆 Lazarus 包
 * @author  wangyouworld
 * @license MIT
 * @site https://ramble.3vshej.cn
 */
class CloneLazarusPackages
{
    /** @var string Lazarus 在线包管理地址 */
    private $cloneUrl;
    /** @var string 在线包管理 json 文件 */
    private $jsonUrl;

    /** @var bool 显示日志 */
    public $showLog = true;

    /** @var array 替换规则 */
    private $replaceRule = [];

    /** @var int 下载缓冲区大小 */
    private $downloadCacheSize = 2 * 1024 * 1024;

    /** @var string 缓存文件名 */
    private $cacheFileName = './packages.cache';

    /**
     * 初始化
     * @throws Exception
     */
    public function __construct($domain = '')
    {
        $domain = $domain ?: 'http://packages.lazarus-ide.org/';
        $this->cloneUrl = $domain;
        $this->jsonUrl = 'packagelist.json';
        if (!substr(php_sapi_name(), 0, 3) == 'cli') {
            throw new Exception('The program runs in CLI mode');
        }
    }

    /**
     * 设置下载缓存区大小
     * @param int $size 设置大小（单位M）
     * @return self
     */
    public function setDownloadCacheSize($size = 1)
    {
        $size = abs($size);
        $size = $size > 0 ? $size : 1;
        $this->downloadCacheSize = $size * 1024;
        return $this;

    }

    /**
     * 设置转换规则
     * <p>$search 说明：</p>
     * <p>null 禁用替换规则</p>
     * <p>string 添加新规则，配合 $replace 使用</p>
     * <p>array 替换规则，下标分别为 search replace，2维数组</p>
     * @param null|string|array $search 搜索项
     * @param string $replace 替换项
     * @return self
     */
    public function setReplaceRule($search = null, $replace = '')
    {
        if ($search === null) {
            $this->replaceRule = [];

            return $this;
        }
        if (is_array($search)) {
            $this->replaceRule = $search;

            return $this;
        }
        $this->replaceRule[] = ['search' => $search, 'replace' => $replace];

        return $this;
    }

    /**
     * 字节转换为人类可读形式
     */
    private function toHumanByte($size, $digits = 2)
    {
        $unit = array('', 'K', 'M', 'G', 'T', 'P'); // 单位数组，是必须1024进制
        $base = 1024;                               // 对数的基数
        $i = floor(log($size, $base));           // 字节数对1024取对数，值向下取整

        return $size ? (round($size / pow($base, $i), $digits) . ' ' . $unit[$i] . 'B') : '0 B';
    }

    /**
     * 写入日志
     * @param string $msg 日志内容
     * @param boolean $force 强制显示
     */
    public function writeLog($msg, $force = false)
    {
        if ($force || $this->showLog) {
            echo $msg, PHP_EOL;
        }
    }

    /**
     * 运行时间计时
     * @return float
     */
    private function startRunTime()
    {
        $mtime1 = explode(" ", microtime());

        return $mtime1[1] + $mtime1[0];
    }

    /**
     * 获取运行时间
     * @param float $start 计时开始时间
     * @param int $s 保留小数位
     * @return float 单位秒
     */
    private function runTime($start, $s = 4)
    {
        $mtime2 = explode(" ", microtime());
        $endTime = $mtime2[1] + $mtime2[0];

        return round($endTime - $start, $s);
    }


    /**
     * 下载文件
     * <p>返回：</p>
     * <p>.local 是否本地文件</p>
     * <p>.size 文件大小</p>
     * @param string $file 文件地址
     * @param string $saveName 保存名
     * @param string $forceDownload 是否强制下载
     * @param boolean $enableReplace 是否应用替换规则（默认为 false）
     * @return array
     * @throws Exception
     */
    public function downloadFile($file, $saveName, $forceDownload = false, $enableReplace = false)
    {
        $aTime = $this->startRunTime();
        $info = '';
        $dirName = dirname($saveName);

        if ($dirName != '.' && !is_dir($dirName)) {

            if (mkdir('.' . $dirName, 0644, true) === false) {
                throw new Exception("create download directory .{$dirName} error");
            }
        }
        if ($forceDownload || !is_file($saveName)) {
            $remoteFileId = fopen($file, 'rb');
            $targetFileId = fopen($saveName, 'wb');
            if ($remoteFileId === false) {
                throw new Exception('open remote file ' . $file . ' error');
            }
            if ($targetFileId === false) {
                throw new Exception('create local file ' . $saveName . ' error');
            }
            $size = 0;
            while (!feof($remoteFileId)) {
                $tmpInfo = fread($remoteFileId, $this->downloadCacheSize);
                if ($tmpInfo === false) {
                    fclose($remoteFileId);
                    fclose($targetFileId);
                    throw new Exception('read remote file ' . $saveName . ' error');
                }
                if ($enableReplace) {
                    $info .= $tmpInfo;
                }
                $isOk = fwrite($targetFileId, $tmpInfo);
                if ($isOk === false) {
                    fclose($remoteFileId);
                    fclose($targetFileId);
                    throw new Exception('cannot write to local file ' . $saveName);
                }
                $size += $isOk ?: 0;
            }
            fclose($remoteFileId);
            fclose($targetFileId);

            if ($enableReplace) {
                foreach ($this->replaceRule as $rep) {
                    $info = str_replace($rep['search'], $rep['replace'], $info);
                }
                $size = file_put_contents($saveName, $info);
                if ($size === null) {
                    throw new Exception('save file ' . $saveName . ' error');
                }
            }
            $res = ['local' => false, 'size' => $size];
        } else {
            $res = ['local' => true, 'size' => filesize($saveName)];
        }
        $bTime = $this->runTime($aTime);
        $this->writeLog(($res['local'] ? 'LOCAL' : 'REMOTE') . ' download file ' . $file . ', filesize ' . ($this->toHumanByte($res['size'])) . ' time ' . $bTime . 's');

        return $res;
    }

    /**
     * 解析并下载资源文件
     * <p>包含 css 、js 脚本本地存储的</p>
     * @param string $file 待解析文件名
     * @return array 需要下载的文件
     */
    private function parseResourceFile($file)
    {
        $content = $this->openFile($file);
        // 解析样式
        $fileList = "/<link.*?href=\"(.+?)\".*?>/";
        preg_match_all($fileList, $content, $file);
        $tmpList = $file[1] ?? [];
        // 解析脚本
        $fileList = "/<script.*?src=\"(.+?)\".*?>/";
        preg_match_all($fileList, $content, $file);
        $tmpList = array_merge($tmpList, $file[1] ?? []);
        $list = [];
        foreach ($tmpList as $v) {
            if (strpos($v, '://') === false) {
                $list[] = $v;
            }
        }
        foreach ($list as $downFile) {
            $this->downloadFile($this->cloneUrl . $downFile, $downFile, true);
        }

        return $list;
    }

    /**
     * 打开本地或远程文件
     * <p>返回全部文件内容</p>
     * @param string $file 文件名
     * @return string
     * @throws Exception
     */
    private function openFile($file)
    {
        $info = file_get_contents($file);
        if ($info === false) {
            throw new Exception('open file ' . $file . ' error');
        }

        return $info;
    }

    /**
     * 是否可以下载包
     * <p>可以下载包：文件不存在、md5 值不同</p>
     * <p>不下载：包 md5 值相同</p>
     * @param string $packageName 包名
     * @param string $md5 md5 值
     * @return bool
     * @throws Exception
     */
    private function isCanDownloadPackage($packageName, $md5)
    {
        if (!is_file($packageName)) {
            return true;
        }
        $checkMd5 = $this->writeCacheFile('get', $md5);
        return strtolower($checkMd5) != strtolower($md5);
    }

    /**
     * 解析 Json 资源文件
     * <p>返回：</p>
     * <p>.data 包信息</p>
     * <p>.files 包文件信息</p>
     * @return array
     * @throws Exception
     */
    private function parseJsonFile($file)
    {
        $content = $this->openFile($file);
        if (!$content) {
            throw new Exception($this->jsonUrl . ' file is empty');
        }
        $list = json_decode($content, true);
        if ($list === null) {
            throw new Exception('parse json file ' . $this->jsonUrl . ' error');
        }
        $packageData = $packageFiles = [];
        foreach ($list as $key => $val) {
            $lbl = strtolower($key);
            if (strpos($lbl, 'packagedata') !== false) {
                $packageData[str_replace('packagedata', '', $lbl)] = $val;
                continue;
            }
            if (strpos($lbl, 'packagefiles') !== false) {
                $packageFiles[str_replace('packagefiles', '', $lbl)] = $val;
                continue;
            }
        }
        if (!$packageData || !$packageFiles) {
            throw new Exception($this->jsonUrl . ' file identify error');
        }

        return ['data' => $packageData, 'files' => $packageFiles];
    }

    /**
     * 下载包
     * <p>返回：</p>
     * <p>.num 文件数</p>
     * <p>.size 下载大小</p>
     * @param array $packageList 包列表
     * @return array
     */
    private function downloadPackage($packageList)
    {
        $download = ['num' => 0, 'size' => 0];

        $this->writeCacheFile('reload');

        $cacheLog = [];

        foreach ($packageList as $file) {
            // 判断文件是否可下载

            if ($this->isCanDownloadPackage($file['RepositoryFileName'], $file['RepositoryFileHash'])) {
                $res = $this->downloadFile($this->cloneUrl . $file['RepositoryFileName'], $file['RepositoryFileName'], true);
                $this->writeCacheFile($file['RepositoryFileHash'], $file['RepositoryFileName'], $file['RepositoryFileSize']);
                $download['num'] += 1;
                $download['size'] += ($res['local'] ? 0 : $res['size']);
            } else {
                $this->writeLog("SKIP\t{$file['RepositoryFileName']}");
            }
            $cacheLog[] = ['md5' => strtolower($file['RepositoryFileHash']), 'name' => $file['RepositoryFileName'], 'size' => $file['RepositoryFileSize']];
        }

        // 重写缓存
        $this->writeCacheFile('reset');
        foreach ($cacheLog as $file) {
            $this->writeCacheFile($file['md5'], $file['name'], $file['size']);
        }

        $this->writeCacheFile('save');

        return $download;
    }

    /**
     * 写入缓存文件
     * <p>命令有：</p>
     * <p>reset 重置</p>
     * <p>reload 重新载入</p>
     * <p>add 添加</p>
     * <p>save 保存</p>
     * <p>get 获取项</p>
     * <p>命令有：</p>
     * @param string $md5 文件 md5（或命令 reset|reload|get|save）
     * @param string $name 包文件名（md5名）
     * @param int $size 包大小
     * @return array|string
     * @throws Exception
     */
    private function writeCacheFile($md5, $name = '', $size = 0)
    {
        static $cacheInfo;
        if (strlen($md5) <= 10) {
            if ($md5 == 'reset') {
                $cacheInfo = ['refresh_time' => date('Y-m-d H:i:s'), 'num' => 0, 'size' => 0, 'list' => []];
                return [];
            }
            if ($md5 == 'reload') {
                $cacheInfo = ['refresh_time' => date('Y-m-d H:i:s'), 'num' => 0, 'size' => 0, 'list' => []];
                if (is_file($this->cacheFileName)) {
                    $info = file_get_contents($this->cacheFileName);
                    if ($info === false) {
                        throw new Exception('open packages cache file error');
                    }
                    $cacheInfo = unserialize($info) ?: $cacheInfo;
                }

                return $cacheInfo;
            }

            if ($md5 == 'get') {
                return isset($cacheInfo['list'][$name]) ? $name : '';
            }

            if ($md5 == 'save') {
                $cacheInfo['refresh_time'] = date('Y-m-d H:i:s');
                $size = file_put_contents($this->cacheFileName, serialize($cacheInfo));
                if ($size === false) {
                    throw new Exception('write packages cache file error');
                }
                $res = $cacheInfo;
                $cacheInfo = [];
                $res['file_size'] = $size;
                return $res;
            }
        }

        if ($md5 && $name && $size) {
            $cacheInfo['num'] += 1;
            $cacheInfo['size'] += $size;
            $cacheInfo['list'][$md5] = $name;
        }

        return [];

    }

    /**
     * 执行克隆操作
     */
    public function clone()
    {
        $aTime = $this->startRunTime();
        $this->writeLog('clone lazarus packages ...');
        // 下载首页文件
        $indexFile = './index.html';
        $this->downloadFile($this->cloneUrl, $indexFile, true, true);

        // 解析并下载资源文件（JS等）
        $this->parseResourceFile($indexFile);
        // 下载 JSON 文件（包文件）
        $this->downloadFile($this->cloneUrl . $this->jsonUrl, $this->jsonUrl, true);
        // 解析 JSON
        $res = $this->parseJsonFile($this->cloneUrl . $this->jsonUrl);
        // 下载包
        $this->writeLog('downloading ...');
        $res = $this->downloadPackage($res['data']);
        $bTime = $this->runTime($aTime);
        $this->showLog = true;
        $this->writeLog("download file:{$res['num']}, filesize:" . $this->toHumanByte($res['size']), true);
        $this->writeLog("run time:{$bTime}s", true);
    }

    /**
     * 验证本地包
     */
    public function checkLocalPackages()
    {
        $this->writeLog('check lazarus packages ...');
        // 下载 JSON 文件
        $this->downloadFile($this->cloneUrl . $this->jsonUrl, $this->jsonUrl);
        // 解析 JSON
        $res = $this->parseJsonFile($this->cloneUrl . $this->jsonUrl);
        $download = ['num' => 0, 'size' => 0, 'update_num' => 0, 'continue_num' => 0, 'download_size' => 0];
        $this->writeLog('diff packages ...');
        $this->writeCacheFile('reload');

        foreach ($res['data'] as $file) {
            // 判断文件是否可下载
            $isDownload = $this->isCanDownloadPackage($file['RepositoryFileName'], $file['RepositoryFileHash']);
            $this->writeLog(($isDownload ? ' YES' : ' NO') . "\t{$file['RepositoryFileName']}");
            if ($isDownload) {
                $download['update_num'] += 1;
                $download['download_size'] += $file['RepositoryFileSize'];
            }
            if (!$isDownload) {
                $download['continue_num'] += 1;
            }
            $download['num'] += 1;
            $download['size'] += $file['RepositoryFileSize'];
        }
        $this->writeLog("packages num:{$download['num']}, total size:" . $this->toHumanByte($download['size']), true);
        $this->writeLog("need update num:{$download['update_num']}, download filesize:" . $this->toHumanByte($download['download_size']) . ", skip num:" . $download['continue_num'], true);
    }

    /**
     * 刷新缓存
     * @throws Exception
     */
    public function refreshCache()
    {
        $this->writeLog('refresh cache ...');
        // 下载 JSON 文件
        $this->downloadFile($this->cloneUrl . $this->jsonUrl, $this->jsonUrl);
        // 解析 JSON
        $res = $this->parseJsonFile($this->cloneUrl . $this->jsonUrl);
        $this->writeLog('check packages ...');

        $this->writeCacheFile('reset');
        foreach ($res['data'] as $file) {
            if (is_file($file['RepositoryFileName'])) {
                $md5 = md5_file($file['RepositoryFileName']);
                $size = filesize($file['RepositoryFileName']);
                if ($md5 === false) {
                    throw new Exception('md5 sum error ' . $file['RepositoryFileName']);
                }
                $md5 = strtolower($md5);

                $this->writeCacheFile($md5, $file['RepositoryFileName'], $size);
            }
        }

        $cacheInfo = $this->writeCacheFile('save');

        $this->writeLog("packages num:{$cacheInfo['num']}, total size:" . $this->toHumanByte($cacheInfo['size']), true);
        $this->writeLog("packages cache time:" . $cacheInfo['refresh_time'], true);
        $this->writeLog("packages cache file size:" . $this->toHumanByte($cacheInfo['file_size']), true);
    }
}

set_time_limit(-1);
ini_set('memory_limit', '128M');

try {

    $command = $argv[1] ?? '';
    $command = in_array($command, ['clone', 'check', 'refresh']) ? $command : '';

    $packageModel = new CloneLazarusPackages();
    $packageModel->showLog = (bool)($argv[2] ?? false);

    switch ($command) {
        case 'check':
            $packageModel->checkLocalPackages();
            break;
        case 'refresh':
            $packageModel->refreshCache();
            break;

        case 'clone':

            // 设置替换规则（将远程引用，改为本地引用）
            $packageModel->setReplaceRule('https://fonts.googleapis.com/css', 'https://ide-packages.freepascal.cn/static/gfonts.css');
            $packageModel->setReplaceRule('https://unpkg.com/bootstrap-material-design@4.1.1/dist/css/bootstrap-material-design.min.css', 'https://ide-packages.freepascal.cn/static/bootstrap-material-design.min.css');
            $packageModel->setReplaceRule('https://ajax.googleapis.com/ajax/libs/angularjs/1.8.0/angular.min.js', 'https://ide-packages.freepascal.cn/static/angular.min.js');

            $packageModel->setReplaceRule('<script src="main.js"></script>', '<script src="https://ide-packages.freepascal.cn/static/main.js"></script>');
            $packageModel->setReplaceRule('<script src="mark.min.js"></script>', '<script src="https://ide-packages.freepascal.cn/static/mark.min.js"></script>');

            $packageModel->setReplaceRule('Lazarus Packages', 'Lazarus 包');
            $packageModel->setReplaceRule('placeholder="Search"', 'placeholder="搜索"');
            $packageModel->setReplaceRule('Download</a>', '下载</a>');
            $packageModel->setReplaceRule('Home Page</a>', '主页</a>');
            $packageModel->setReplaceRule('Package:', '包名：');
            $packageModel->setReplaceRule('Author:', '作者：');
            $packageModel->setReplaceRule('Description:', '介绍：');
            $packageModel->setReplaceRule('License:', '授权：');
            $packageModel->setReplaceRule('Dependencies:', '依赖：');
            $packageModel->setReplaceRule('Version:', '版本：');

            $packageModel->clone();
            break;
        default:
            $packageModel->writeLog('Clone Lazarus Packages V0.5', true);
            $packageModel->writeLog('', true);
            $packageModel->writeLog('Usage: php CloneLazarusPackages.php [clone|check|refresh] [log]', true);
            $packageModel->writeLog('', true);
            $packageModel->writeLog("clone\tclone lazarus packages site", true);
            $packageModel->writeLog("check\tcheck packages", true);
            $packageModel->writeLog("refresh\trefresh packages md5 cache", true);
            $packageModel->writeLog('', true);
            $packageModel->writeLog("log\tshow log", true);
            $packageModel->writeLog('', true);
            $packageModel->writeLog('Examples:', true);
            $packageModel->writeLog("php CloneLazarusPackages.php clone log", true);
            $packageModel->writeLog("php CloneLazarusPackages.php check log", true);
            $packageModel->writeLog("php CloneLazarusPackages.php refresh", true);

            break;
    }
} catch (Exception $e) {
    echo $e->getMessage(), PHP_EOL;
    // echo $e->getTraceAsString();
}
